/**
 * @file   bindex_table.c
 * @author David Bremner
 * @date   Tue Mar 26 15:32:52 2002
 * 
 * @brief  2-table indexed by basis indices. Each entry in the table consists 
 * of a fixed width sequence of indices. If you like a three dimensional array.
 * 
 * Current use is to store facets adjacent to two other specified facets.
 * 
 */

#include <stdlib.h>
#include "config.h"
#include "nalloc.h"
#include "basis_equation.h"
#include "bindex_table.h"
#include "oms_error.h"
#include "combinat.h"

static void find_bridge_bases(bindex_table_p,size_t,size_t,be_block_p);


/** 
 * Allocate a table from the heap; not sure why you want to.
 * 
 * @param rows 
 * @param cols 
 * @param depth 
 * 
 * @return 
 */
extern bindex_table_p bt_alloc(size_t rows, 
			   size_t cols, 
			   size_t depth){
  bindex_table_p new;

  new=malloc(sizeof(bindex_table_t)+depth*rows*cols*sizeof(array_index_t));
  TRAP_NULL(new);

  new->rows=rows;
  new->cols=cols;
  new->depth=depth;

  return (new);
}

extern na_node_p bt_nalloc(na_node_p node,
			   array_index_t slot,
			   size_t rows, 
			   size_t cols, 
			   size_t depth){
  bindex_table_p new;

  node=na_alloc(node,slot,sizeof(bindex_table_t)+rows*cols*depth*sizeof(array_index_t));

  new=na_access(node,slot);
  TRAP_NULL(new);

  new->rows=rows;
  new->cols=cols;
  new->depth=depth;

  return (node);
}

/** 
 * Build table of bridges, i.e.\ facets that are adjacent to two specified
 * facets.
 * 
 * @param BT table of the appropriate size to be initialized
 * @param n number of elements
 * @param r rank of om.
 * @param pivots table of neigbours.
 */
extern void bt_bridges_init(bindex_table_p BT,
			    natural_t n,
			    natural_t r,
			    be_block_p pivots){
  natural_t num_bases;

  num_bases=binomial(n,r-1);

  oms_assert(BT->cols==num_bases);
  oms_assert(BT->rows==num_bases);
  oms_assert(BT->depth==(n-1));
  find_bridge_bases(BT,n,r,pivots);

}



/**
   find the facets adjacent to each pair of facets. Note that
   there could be none, or several.

*/

static void find_bridge_bases(bindex_table_p BT,size_t n, size_t r,
			      be_block_p pivots){
  int i,j;

  array_index_t *i_nbr,*j_nbr;

  i_nbr=calloc(be_cols(pivots),sizeof(array_index_t));
  j_nbr=calloc(be_cols(pivots),sizeof(array_index_t));

  for (i=0; i< be_rows(pivots); i++){
    for (j=i+1; j< be_rows(pivots); j++){
      int k, i_off, j_off,equal;
      /* find neighbours common to i,j */
      for (k=0; k<be_cols(pivots); k++){
	i_nbr[k]=be_bindex(pivots,i,k);
	j_nbr[k]=be_bindex(pivots,j,k);
      }
      qsort(i_nbr, be_cols(pivots), sizeof(array_index_t), index_cmp);
      qsort(j_nbr, be_cols(pivots), sizeof(array_index_t), index_cmp);

      i_off=0;
      j_off=0;
      equal=0;

      while (i_off < be_cols(pivots) && 
	     j_off < be_cols(pivots)){
	if (i_nbr[i_off] == j_nbr[j_off]){
	  oms_assert(equal<BT->depth);

	  BT_ELEM(BT,i,j,equal)=i_nbr[i_off];

	  equal++;
	  i_off++;
	  j_off++;
	}
	  
	if (i_nbr[i_off] < j_nbr[j_off]){
	  i_off++;
	} else if (i_nbr[i_off] > j_nbr[j_off]){
	  j_off++;
	}
      }

      BT_ELEM(BT,i,j,equal)=-1; /* mark the last one */

    }
    



  }
  free(i_nbr);
  free(j_nbr);

      
}
