#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "nodedef.h"
#include "basis_equation.h"
#include "nodedef.h"
#include "globaldef.h"
#include "oms_opt.h"
#include "profile.h"
#include "bitmask.h"
#include "debug.h"
#include "combinat.h"
#include "boundary.h"

static void find_boundary_bases(be_block_p be,size_t,size_t);

static array_index_t basis_extensions(be_block_p BE,
				   array_index_t row,
				   array_index_t col,
				   natural_t n, natural_t r,
				   natural_t *pi,
				   array_index_t forbidden);


static void find_ridge_table(be_block_p,size_t,size_t);

static void find_neighbour_bases(be_block_p,size_t,size_t);


/**
 * Find bases for boundary equations.

 *
 * Here we take advantage of the fact that the basis indexing is 
 * in lexicographic order. If we have a facet \e r-1 basis  \em F, 
 * then the bases containing it are <em>{F,a[1]} ... {F,a[n-r]}</em> where
 * the \em a[i]'s are vertices not occuring in \e F. 
 */
static void find_boundary_bases(be_block_p bdy,size_t n, size_t r){
  int j;

  natural_t *pi;
  natural_t *pibar;
  int *tau;
  int d=r-1;
  int row=0;

  pi=calloc(d+1,sizeof(natural_t));
  pibar=calloc(n-d,sizeof(natural_t));
  tau=calloc(n-d,sizeof(int));

  first_kpartition(n,d,pi,pibar,tau);
  

  do{
    for (j=0; j< n-d; j++){
      be_set_sign(bdy,row,j,tau[j]);
      swap(pi[d],pibar[j]);
      be_set_bindex(bdy,row,j,kset_cindex(n,r,pi));
      swap(pi[d],pibar[j]);
    }
    row++;
  } while(next_kpartition(n,d,pi,pibar,tau));

  free(pi);
  free(pibar);
  free(tau);
	  
	  
}



/** 
 * Traverse boundary equations.

 * This version provides a hook to be called for each forced facet.
 * It requires a two more state arrays, pos_mask and neg_mask to 
 * be defined.


@param  bindex  what basis to check
@param	local	backtracking search state
@param	global	global variables (read only)
@param	Q	Queue of basis indices.
- input: first element -- variable to traverse
- output: bases forced by this traversal
@param	force_hook function to call for each facet forced onto the boundary.
*/
bool be_boundary_traverse2(array_index_t bindex,
			    na_node_p local, 
			    na_node_p global,
			    signed_set_p S,
			    oms_hook force_hook) {
  int elem, chi_bindex;
  be_block_p BE,pivots;
  sarray_p chi, pos_mask,neg_mask, active;
  ridge_graph_p RG;

  chi=na_access(local,SLOT_chirotope);
  BE=na_access(global,SLOT_bdy_eq);
  pos_mask=na_access(local,SLOT_pos_mask);
  neg_mask=na_access(local,SLOT_neg_mask);
  RG=na_access(local,SLOT_ridge_graph);
  pivots=na_access(global,SLOT_pivots);
  active=na_access(local,SLOT_boundary);

  chi_bindex=sa_read(chi,bindex);

  elem=BE->last[bindex];   
  err_assert( elem >= -1 && elem < BE->rows*BE->cols );

  while ( elem>=0  ){
    array_index_t row,col,i;
    int32_t pmask,nmask,sigma,missing_bit;
    int row_active;
    int32_t new_bit;    

    row=be_row(BE,elem);
    col=be_col(BE,elem);

    row_active=sa_read(active,row);

    err_debug(DB_BOUNDARY,("be_boundary_traverse2","row=%d, active=%d",row,row_active));

    new_bit=(1 << col);

    sigma=chi_bindex * be_sign_raw(BE,elem);
    err_assert(sigma*sigma==1);

    if (sigma==1){
      pmask=sa_or(pos_mask,row,new_bit);
      nmask=sa_read(neg_mask,row);
    } else {
      pmask=sa_read(pos_mask,row);
      nmask=sa_or(neg_mask,row,new_bit);
    }
    
    if (ERR_DEBUG_MASK & DB_BOUNDARY)
      err_assert(be_sanity_check_boundary(BE,row,chi,pos_mask,neg_mask));

    err_assert((pmask & nmask)==0);
    err_debug(DB_BOUNDARY,("be_boundary_traverse2","row=%d, mask=%x %x",row,nmask,pmask));

    switch (row_active){
    case 0:
      /* Currently Undecided */
      if ( (pmask|nmask) == BE->ones_mask ){
	if ((pmask== BE->ones_mask) || (nmask == BE->ones_mask)){
	  err_debug(DB_BOUNDARY|DB_FORCE,("be_boundary_traverse2", "forcing %d onto boundary",row));
	  maybe_prof_event(PROF_BOUNDARY_FORCED);

		
	  sa_set(active,row,1);
	  if (force_hook && !force_hook(local,global,row,1))
	    return FALSE;
	  
	  if (BE->diameter !=-1){
	    if (!rg_add_facet(RG,row,active,pivots))
	      ERR_RETURN(FALSE);
	  }
	} else {
	  err_debug(DB_BOUNDARY|DB_FORCE,("be_boundary_traverse2", "forcing %d off boundary",row));

	  sa_set(active,row,-1);
	}
      }
      break;
      
    case 1:
      /* Already on the boundary */
      if ( (pmask == 0 ) || nmask==0 ){
	if ((pmask|nmask) != BE->ones_mask){
	  for (i=be_row_start(BE,elem); 
	       i<be_row_start(BE,elem)+BE->cols; i++){
	    array_index_t b=be_bindex_raw(BE,i);
	    int chi_i;

	      err_assert(b >=0 && b <= sa_size(chi));
	      if (b != bindex){
	      
		chi_i=sa_read(chi,b)*be_sign_raw(BE,i);
		if (chi_i==0){
		  int new=sigma*be_sign_raw(BE,i);
		  err_debug(DB_BOUNDARY,("be_boundary_traverse2", "forcing %d to %d",
			       b,new));
		  if (!ss_add(S,b,new)) return FALSE;
		} else {
		  if (chi_i != sigma) ERR_RETURN(FALSE);
		}
	      }
	    }
	}
      } else {
	ERR_RETURN(FALSE);
      }
      break;

    case -1:      /* Not permitted on the boundary */
      /*\todo this looks wrong. check? */
      if ( (pmask|nmask)==BE->ones_mask){
	if ((pmask==0) || (nmask==0)){
	  err_debug(DB_RETURNS|DB_BOUNDARY,(HERE,"row %d contradicted\n",row));
	  ERR_RETURN(FALSE);
	} else {
	  missing_bit= (~(pmask|nmask))&BE->ones_mask;
	  if ( ZERO_OR_POWER_OF_TWO(missing_bit)  && ((pmask==0) || (nmask==0))){
	    sign_t new_sign;
	    array_index_t basis,array_index_t;
	    
	    err_debug(DB_BOUNDARY,("be_boundary_traverse2", "pm %d nm %d mb %d ",
			 pmask,nmask,missing_bit));

	    err_assert(missing_bit && ZERO_OR_POWER_OF_TWO(missing_bit));


	    array_index_t=be_row_start(BE,elem)+LOG2(missing_bit);
	    basis=be_bindex_raw(BE,array_index_t);

	    new_sign=sigma*(-1)*be_sign_raw(BE,array_index_t);
	    

	    err_debug(DB_BOUNDARY|DB_FORCE,("be_boundary_traverse2", "not facet %d forcing %d (%d) to %d",
			 row, basis,missing_bit,new_sign));

	    if (!ss_add(S,basis,new_sign)) return FALSE;
	  }
	}
      }
      break;
    default:
      err_fatal(HERE,"Unknown row_active value %d",row_active);
    } /* switch (row_active) */

    elem=BE->eqn[elem].prev;  
  } /* while (elem>0) */

  ERR_RETURN(TRUE);
}

/** 
 * Initialize an appropriately sized basis equation block for the 
 * boundary equations.
 */

extern void be_boundary_init(be_block_p bdy_eq, natural_t n, natural_t r,
			      array_index_t source,array_index_t sink,array_index_t diameter){
  natural_t num_bases,num_eqns;

  num_bases=binomial(n,r);
  num_eqns=binomial(n,r-1);

  err_assert(bdy_eq->cols==n-r+1);
  err_assert(bdy_eq->bases==num_bases);
  err_assert(bdy_eq->rows==num_eqns);
  err_assert(bdy_eq->eqn !=NULL);

  bdy_eq->stride=1;
  bdy_eq->source=source;
  bdy_eq->sink=sink;
  bdy_eq->diameter=diameter;
  find_boundary_bases(bdy_eq,n,r);

  be_makelinks(bdy_eq);  
}




static void find_neighbour_bases(be_block_p BE,size_t n, size_t r){
  natural_t *pi;
  int row=0;

  pi=calloc(r+1,sizeof(natural_t));

  first_kset(n,r-1,pi+1);
  
  do{
    array_index_t leave;
    natural_t resting=0;
    array_index_t col=0;
    
    for (leave=1; leave < r ; leave++){
      swap(pi[leave],resting);
      pi[r]=n+1;
      pi[0]=1;
      col=basis_extensions(BE,row,col,n,r-1,pi,resting);
      swap(pi[leave],resting);
    }
    row++;
  } while(next_kset(n,r-1,pi+1));
  free(pi);
}


static array_index_t basis_extensions(be_block_p BE,
				   array_index_t row,
				   array_index_t col,
				   natural_t n, 
				   natural_t r,
				   natural_t *pi,
				   array_index_t forbidden ){
  int i,sign,skipped;

  err_assert(n>=r); 
  err_assert(pi!=NULL); 
  err_assert(BE !=NULL);

  i=0;

  skipped=0;
  sign=1;
    
  while(i-skipped<r && pi[i]<=n ){
    if (pi[i]==forbidden) {
      pi[i]++;
    } else {
      if (pi[i]==pi[i+1] ) {
	pi[i]++;
      } 
      
      if(pi[i] > pi[i+1]){
	swap(pi[i],pi[i+1]);
	if (pi[i] !=0) {
	  sign *= -1;
	}	else {
	  skipped++;
	}
	i++;
      } else {
	be_set_bindex(BE,row,col,kset_index(n,r,pi));
	be_set_sign(BE,row,col,sign);
	col++;
	pi[i]++;
      }
    }
  }
  /*    move things back over  */
  for(i=r; i>0; i--)
      pi[i]=pi[i-1];
	
  return col;
}

/** 
   Construct a table where each row is indexed by a facet, and
   each row contains the basis indices of all of the ridges 
   contained in that facet.
*/


static void find_ridge_table(be_block_p ridge_table, size_t n, size_t r){

  natural_t *pi;
  natural_t resting=0;
  array_index_t leave,beta;
  
  pi=calloc(r,sizeof(natural_t));

  first_kset(n,r-1,pi);
  beta=0;
  do{
    for (leave=0; leave < r-1 ; leave++){
      swap(pi[leave],resting);
      be_set_bindex(ridge_table,beta,leave,kset_index(n,r-2,pi));
      swap(pi[leave],resting);
    }
    beta++;
  } while(next_kset(n,r-1,pi));


  free(pi);
}





extern bool be_sanity_check_boundary(be_block_p BE, array_index_t row, 
			  sarray_p chi, 
			  sarray_p pos_mask,
			  sarray_p neg_mask){


  int32_t pmask,nmask;
  be_coeff_p Eqn;
  int j,row_start;

  pmask=sa_read(pos_mask,row);
  nmask=sa_read(neg_mask,row);

  row_start=(BE->cols)*row;

  Eqn=BE->eqn+row_start;

  for (j=0; j<BE->cols; j++){
    int32_t new_bit;    
    int sigma;
   
    new_bit = (1 << j);
    sigma=be_sign_raw(BE,row_start+j) * sa_read(chi,
						be_bindex_raw(BE,row_start+j));

    
    switch (sigma){
    case 0:
      if ((pmask|pmask) & new_bit)
	return FALSE;
      break;
    case 1:
      if (nmask & new_bit)
	return FALSE;
      if (!(pmask & new_bit))
	return FALSE;
      break;
    case -1:
      if (!(nmask & new_bit))
	return FALSE;
      if ((pmask & new_bit))
	return FALSE;
      break;
    default:
      err_fatal(HERE,"bizarre sign value %d",sigma);
    }
  }
  return TRUE;
}

    

/**
 Initialize a table to compute ridges of a facet 
 */
extern void be_ridge_table_init(be_block_p BE, natural_t n, natural_t r){
  natural_t num_bases,num_eqns;

  num_bases=binomial(n,r-2);
  num_eqns=binomial(n,r-1);

  err_assert(BE->cols==(r-1));
  err_assert(BE->bases==num_bases);
  err_assert(BE->rows==num_eqns);
  err_assert(BE->eqn !=NULL);

  BE->stride=1;

  find_ridge_table(BE,n,r);


  be_makelinks(BE);  
}


/**
 * Initialize a table to compute possible neigbours in the boundary 
 * complex 
 */
extern void be_neighbour_init(be_block_p BE, natural_t n, natural_t r){
  natural_t num_bases,num_eqns;

  num_bases=binomial(n,r-1);
  num_eqns=binomial(n,r-1);

  err_assert(BE->cols==(n-r+1)*(r-1));
  err_assert(BE->bases==num_bases);
  err_assert(BE->rows==num_eqns);
  err_assert(BE->eqn !=NULL);

  BE->stride=1;

  find_neighbour_bases(BE,n,r);


  be_makelinks(BE);  
}

