/**
 * @file   variable_choice.c
 * @author David Bremner <bremner@unb.ca>
 * @date   Thu Jul  4 15:05:56 2002
 * 
 * @brief  heuristics for chirotope backtracking variable choice
 * 
 * 
 */

#include "nodedef.h"
#include "globaldef.h"
#include "oms_error.h"
#include "oms_types.h"
#include "statearray.h"
#include "nalloc.h"
#include "basis_equation.h"
#include "ridge_graph.h"
#include "profile.h"
#include "uni.h"
#include "combinat.h"
#include "nu_sign_set.h"
#include "pluecker.h"

static int vc_score(na_node_p global, na_node_p state, array_index_t basis);



/** 
 * Choose the first unset variable in some static order
 * 
 * 
 * @return unset variable, or 0 if none are found
 */
array_index_t vc_static(sarray_p chirotope,
		  array_index_t *reorder_table ){
  int i;

  for (i=0; i<sa_size(chirotope); i++){
    int j;
    j=reorder_table[i];
    if (sa_read(chirotope,j)==0)
      return si_encode(j,1);
  }
  return 0;
}

/** 
 * choose a variable that (hopefully) forces the state of a facet.
 * 
 * @param global global state
 * @param local local state
 * @param order 
 * 
 * @return variable choice, or 0 if none.
 */
array_index_t vc_facet(na_node_p global, na_node_p local, int order){
  
  ridge_graph_p RG=na_access(local,SLOT_ridge_graph);
  be_block_p bdy_eq=na_access(global,SLOT_bdy_eq);
  sarray_p chirotope=na_access(local,SLOT_chirotope);
  sarray_p boundary=na_access(local,SLOT_boundary);
  
  array_index_t best_score=0,best_sign=1,best_facet=-1;
  sign_t sign_wanted=1;
  
  int i,j;

  if (sa_zeros(chirotope)==0)
    return 0;

/*   score[0]=calloc(sa_size(chirotope),sizeof(natural_t)); */
/*   score[1]=calloc(sa_size(chirotope),sizeof(natural_t)); */
  
  for (i=0; i<RG->bases; i++){
    /* find a fringing facet */
    if ((sa_read(RG->source_dist,i) < RG_INFINITY) && (sa_read(boundary,i)==0)){
      int row_non_zeros=0;
      int score=0;

      sign_wanted=1;

      for (j=0; j < bdy_eq->cols; j++){
	sign_t sigma=sa_read(chirotope,be_bindex(bdy_eq,i,j));
 	if (sigma != 0){ 
 	  sign_wanted=sigma*be_sign(bdy_eq,i,j); 
 	  row_non_zeros++; 
	}  
      }

      score=row_non_zeros
 	+sa_read(RG->source_dist,i) 
 	+sa_read(RG->sink_dist,i); 
      if (score > best_score){
	best_facet=i;
	best_score=score;
	best_sign=sign_wanted;
      }
    }
  }

  oms_assert(best_facet>=0);
  oms_debug(3,("vc_facet", "trying facet %d to %d\n",best_facet,best_sign));

  for (j=0; j < bdy_eq->cols; j++){
    array_index_t beta=be_bindex(bdy_eq,best_facet,j);
    sign_t coeff=be_sign(bdy_eq,best_facet,j);
    sign_t sigma=sa_read(chirotope,beta);
    if (sigma == 0){
      oms_debug(3,("vc_facet", "trying basis %d to %d\n",beta,best_sign*coeff));      
      return si_encode(beta,best_sign*coeff);
    }
  }
  return -1;
/*   for (i=0; i<sa_size(chirotope); i++){ */
/*     for (j=0; j<2; j++){ */
/*       if (score[j][i] > best_score){ */
/* 	best_var=i; */
/* 	best_sign=2j-1; */
/* 	best_score=score[j][i]; */
/*       } */
/*     } */
/*   } */
/*   free(score[0]); */
/*   free(score[1]); */
/*   return si_encode(best_var,best_sign); */
      
}

array_index_t vc_greedy(na_node_p global, na_node_p local, int order){
  int i, best_var=-1, best_score=-1;
  sarray_p chirotope=na_access(local,SLOT_chirotope);
  for (i=0; i < sa_size(chirotope); i++){
    if (sa_read(chirotope,i) == 0) {
      int score=vc_score(global,local,i);
      oms_assert(score>0);
      if (score > best_score){
	best_score=score;
	best_var=i;
      }
    }
  }

  return si_encode(best_var,1);

}
  

/**
   \brief Give variable a score as a branching variable.

   \param basis Which variable to evaluate
*/

static int vc_score(na_node_p global, na_node_p state, array_index_t basis){

  be_block_p PE;		/* %%Pl\"ucker%% equations */
  sa16_array_p pe_state;
  sarray_p chirotope;
  signed_set_p S;		/* variables set, but not yet processed */
  params_p params;
  bool success;
  int score;


  chirotope=na_access(state,SLOT_chirotope);
  pe_state=na_access(state,SLOT_pe_state);
  PE=na_access(global,SLOT_pe);
  params=na_access(global,SLOT_params);

  switch (params->variable_order){
  case 'r':
    return 1+uni()*sa_size(chirotope);
    break;
  case 'g':
    S=ss_new(sa_size(chirotope));
    
    sa_save(chirotope);
    sa16_save(pe_state);

    sa_set(chirotope,basis,1);

    success=pl_traverse(basis,PE,S,chirotope,pe_state);
    if (!success){
      /* failure is good */
      score=sa_size(chirotope);
    } else {
      score=ss_count(S);
    }
    ss_destroy(S);
    sa16_restore(pe_state);
    sa_restore(chirotope);

    oms_debug(4,(HERE,"score=%d",score));
    return score;
    break;
  default:
    oms_fatal(HERE,"unknown variable order %c\n",params->variable_order);
  }

  /*NOTREACHED*/
  return 0;
}

/** Compute an "incremental" order of bases where all bases involving 1..k
    occur before those involving k+1
*/
extern void vc_init_reorder_table(int order, 
			    natural_t n, natural_t r, array_index_t *table){

  natural_t *pi,*pibar;
  sign_t *tau;
  int i,j;

  switch(order){
  case 'i':
    pi=calloc(r,sizeof(natural_t));
    pibar=calloc(n-r,sizeof(natural_t));
    tau=calloc(n-r,sizeof(sign_t));

    j=0;
    for (i=r; i<=n ; i++){
      first_kpartition(i-1,r-1,pi,pibar,tau);
      pi[r-1]=i;
      do {
	table[j++]=kset_cindex(n,r,pi);
      } while (next_kpartition(i-1,r-1,pi,pibar,tau));
    }
    break;
  case 'L':
    for (i=0; i<binomial(n,r); i++)
      table[i]=binomial(n,r)-i-1;
    break;
  case 'l':
  default:
    for (i=0; i<binomial(n,r); i++)
      table[i]=i;
    break;
  }
}
/** 
 * Decide which chirotope variable to branch on next, 
 * and what sign to choose first.
 * @param global global state
 * @param S search node
 * 
 * @return signed basis index, 0 if none unset
 */
signed_index_t vc_choose_unset(na_node_p global,na_node_p S){
  
  sarray_p chirotope;
  params_p params;
  nu_sign_set_p choice_Q;
  array_index_t *reorder_table;


#ifdef USE_PROFILE
  prof_event(PROF_VAR_CHOICE);
#endif

  params=na_access(global,SLOT_params);
  chirotope=na_access(S,SLOT_chirotope);
  choice_Q=na_access(S,SLOT_choice_Q);

  while (!nss_empty(choice_Q)){
    array_index_t k=nss_deq_index(choice_Q);
    
    if (sa_read(chirotope,k)==0){
      oms_debug(3,("ChooseUnset","choosing %d from Q",k));
      return si_encode(k,1);
    }
  }
#ifdef USE_PROFILE
  prof_event(PROF_FREE_CHOICE);
#endif
  
  switch(params->variable_order){
  case 'f':
    return vc_facet(global,S,params->variable_order);
    break;
  case 'l':
  case 'L':
  case 'i':
    reorder_table=na_access(global,SLOT_reorder_table);
    return vc_static(chirotope, reorder_table );
    break;
  case 'r':
  case 'g':
    return vc_greedy(global,S,params->variable_order);
    break;
  }
  /*NOTREACHED*/
  return 0;
}

